import { Component, Input, Output, EventEmitter, OnChanges, OnDestroy } from "@angular/core";

import { NgParameter } from "../../formulaire/ngparam";
import { ServiceFactory } from "../../services/service-factory";
import { ParamValueMode, Observer, ParamDefinition } from "jalhyd";
import { FormulaireService } from "../../services/formulaire/formulaire.service";
import { FormulaireDefinition } from "../../formulaire/definition/form-definition";

@Component({
    selector: "param-link",
    templateUrl: "./param-link.component.html"
})
export class ParamLinkComponent implements OnChanges, Observer, OnDestroy {
    // paramètre géré (qui sera lié à une valeur, cad qui importe cette valeur)
    @Input("param")
    private _param: NgParameter;

    @Output()
    private onValid: EventEmitter<boolean>;

    /**
     * indice actuel du paramètre sélectionné dans la lsite
     */
    private _currentIndex = -1;

    /**
     * message affiché à côté du select des paramètres
     */
    private _message: string;

    /**
     * liste des paramètres liables sous la forme
     * {"name":<étiquette>, "value":<valeur liable>, "nub":<Nub d'origine du paramètre>, "formTitle":<nom de la calculette liée au nub>}
     * 
     * l'étiquette "name" est de la forme <n|N1>[.[N2]]
     * n : indice de de l'ouvrage dans le cas des ouvrages parallèles
     * N1 : un nom de paramètre/résultat (dans le cas d'un résultat, il est suivi d'un point)
     * N2 : nom de résultat complémentaire (optionnel)
     * ex : 
     *   Q, Z1 (paramètres)
     *   J. (résultat)
     *   .Yf (résultat complémentaire du résultat courant)
     *   Q.Yf (résultat complémentaire du résultat nommé "Q")
     */
    private _linkableParams: any[];

    private _formService: FormulaireService;

    constructor() {
        this.onValid = new EventEmitter();
        this._formService = ServiceFactory.instance.formulaireService;
        this._formService.addObserver(this);
    }

    /**
     * envoi d'un événement de validité
     */
    private emitValidity() {
        // this.onValid.emit(this._validList);
    }

    /**
     * réception d'un événement du menu des paramètres liables
     */
    private onSelectLinkableParam(event: any) {
        const next = event.target.value;

        let i = 0;
        for (const e of this._linkableParams)
            if (this._linkableParams[i].value.uid == next.value.uid) {
                this.linkTo(i);
                break;
            }
            else
                i++;
    }

    /**
     * valeur courante affichée dans le select des paramètres liables
     */
    private get currentLinkedParamLabel(): string {
        if (this._linkableParams !== undefined) {
            if (this._currentIndex === -1 || this._currentIndex >= this._linkableParams.length)
                return undefined;

            return this.selectItemLabel(this._linkableParams[this._currentIndex]);
        }

        return undefined;
    }

    //     // le paramètre est il déjà lié à une valeur ? si oui laquelle ?
    //     if(this._currentIndex === -1 && this._param.valueMode == ParamValueMode.LINK && this._linkableParams !== undefined) {
    //     let i = 0;
    //     for (const e of this._linkableParams)
    //         if (e.param.uid === this._param.paramDefinition.re) {
    //             this._currentIndex = i;
    //             break;
    //         }
    //         else
    //             i++;
    // }

    /**
     * attribut "label" d'une entrée du select des paramètres
     */
    private selectItemLabel(i: any) {
        const s = i.name; // nom associé au paramètre/à la valeur
        const c = i.formTitle; // nom de la calculette

        const re5 = /(\d+)\.(.+)\.$/;  // forme <nombre>.xxx. (résultat d'ouvrage)
        const match5 = re5.exec(s);
        if (match5 !== null) {
            const n = +match5[1] + 1
            return `${match5[2]} (résultat de ${c}, ouvrage n°${n})`;
        }

        const re1 = /([^\.]+)\.$/;  // forme xxx. (résultat)
        const match1 = re1.exec(s);
        if (match1 !== null)
            return `${match1[1]} (résultat de ${c})`;

        const re4 = /(\d+)\.(.+)/;  // forme <nombre>.xxx (ouvrage)
        const match4 = re4.exec(s);
        if (match4 !== null) {
            const n = +match4[1] + 1
            return `${match4[2]} (${c}, ouvrage n°${n})`;
        }

        const re2 = /([^\.]+)\.(.+)/;  // forme xxx.yyy (résultat complémentaire)
        const match2 = re2.exec(s);
        if (match2 !== null)
            return `${match2[2]} (${c}, résultat complémentaire de ${match2[1]})`;

        const re3 = /^\.(.+)/;  // forme .xxx (résultat complémentaire)
        const match3 = re3.exec(s);
        if (match3 !== null)
            return `${match3[1]} (${c}, résultat complémentaire)`;

        return `${s} (${c})`; // forme simple (paramètre)
    }

    /**
     * lie le paramètre géré à un des paramètres liables de la liste
     * @param index indice dans la liste
     */
    private linkTo(index: number) {
        if (this._currentIndex !== index) {
            this._currentIndex = index;
            const lp = this._linkableParams[index];

            this._param.linkToParameter(lp.nub, lp.name);
        }
    }

    public updateParamList() {
        // liste des paramètres liables

        if (this._param.valueMode === ParamValueMode.LINK)
            this._linkableParams = this._formService.filterLinkableValues(this._formService.getLinkableValues(this._param));
        else
            this._linkableParams = [];

        // initialisation de l'indice courant

        if (this._linkableParams.length > 0) {
            if (this._currentIndex === -1)
                this.linkTo(0);
            this._message = undefined;
        }
        else {
            this._currentIndex = -1;
            this._message = "Aucun paramètre compatible trouvé";
        }
    }

    public ngOnChanges() {
        if (this._param !== undefined)
            this._param.removeObserver(this);
        this._param.addObserver(this);

        this.updateParamList();
    }

    public ngOnDestroy() {
        this._param.removeObserver(this);
    }

    // interface Observer 

    public update(sender: any, data: any) {
        if (sender instanceof FormulaireService) {
            switch (data["action"]) {
                case "createForm":
                    this.updateParamList();
                    break;
            }
        }
    }
}
