import { Component, ViewChild } from "@angular/core";

import { ArrayReverseIterator, ResultElement, NumberIterator } from "jalhyd";

import { InternationalisationService } from "../../services/internationalisation/internationalisation.service";
import { LogComponent } from "../../components/log/log.component";
import { RemousResults } from "../../results/remous-results";
import { CalculatorResults } from "../../results/calculator-results";
import { VarResultsComponent } from "../fixedvar-results/var-results.component";

@Component({
    selector: "remous-results",
    templateUrl: "./remous-results.component.html",
    styles: [`
    .result_label {
        text-align: right;
        padding-top:10px;
        padding-bottom:10px;
        padding-right:10px;
    }
    .result_value {
        text-align: center;
        padding-left:30px;
        padding-right:30px;
    }
    .result_id_0 {
        background-color: #f0f0f0;
    }
    .result_id_2 {
        font-weight: bold;
    }
    `
    ]
})
export class RemousResultsComponent {
    private _remousResults: RemousResults;

    /*
    * config du graphe principal
    */
    private graph1_type = "line";
    private graph1_data = {};
    private graph1_options = {};

    /*
    * config du graphe secondaire (paramètre supplémentaire sauf Hs, Hsc, Yf, Yt, Yco)
    */
    private graph2_type = "line";
    private graph2_data = {};
    private graph2_options = {};

    /**
     * entêtes de la table de résultats numériques
     */
    private _tableHeaders: string[];

    /**
     * true si les résultats doivent être mis à jour
     */
    private _doUpdate = false;

    /**
     * composant des résultats variables
     */
    @ViewChild(VarResultsComponent)
    private varResultsComponent: VarResultsComponent;

    /**
     * composant journal
     */
    @ViewChild(LogComponent)
    private logComponent: LogComponent;

    constructor(private intlService: InternationalisationService) {
    }

    private get uitextLigneFluviale() {
        return this.intlService.getExtraResLabel("FLU");
    }

    private get uitextLigneTorrentielle() {
        return this.intlService.getExtraResLabel("TOR");
    }

    private get uitextAbscisse() {
        return this.intlService.localizeText("INFO_REMOUSRESULTS_ABSCISSE");
    }

    private get uitextTirant() {
        return this.intlService.localizeText("INFO_REMOUSRESULTS_TIRANT");
    }

    private get uitextFond() {
        return this.intlService.localizeText("INFO_REMOUSRESULTS_FOND");
    }

    private get uitextBerge() {
        return this.intlService.localizeText("INFO_REMOUSRESULTS_BERGE");
    }

    private get uitextTirantNormal() {
        return this.intlService.localizeText("INFO_REMOUSRESULTS_TIRANTNORMAL");
    }

    private get uitextTirantCritique() {
        return this.intlService.localizeText("INFO_REMOUSRESULTS_TIRANTCRITIQUE");
    }

    private get extraGraph(): boolean {
        return this._remousResults === undefined ? false : this._remousResults.extraGraph;
    }

    private get extraParamLabel(): string {
        return this._remousResults === undefined ? undefined :
            this.intlService.getExtraResLabel(this._remousResults.extraParamSymbol);
    }

    public set results(rs: CalculatorResults[]) {
        this._remousResults = undefined;
        if (rs !== undefined) {
            for (const r of rs) {
                if (r instanceof RemousResults) {
                    this._remousResults = r;
                    break;
                }
            }
        }
        this.updateView();
    }

    public updateView() {
        this.graph1_data = {};
        this.graph1_options = {};
        this.graph2_data = {};
        this.graph2_options = {};
        if (this.varResultsComponent !== undefined) {
            this.varResultsComponent.results = undefined;
        }
        if (this.logComponent !== undefined) {
            this.logComponent.log = undefined;
        }
        this._tableHeaders = [];

        if (this._remousResults !== undefined) {
            this._doUpdate = this._remousResults.hasResults;
        }
    }

    /**
     * appelé pour gérer les changements non détectés par Angular
     */
    public ngDoCheck() {
        if (this._doUpdate) {
            this._doUpdate = !this.updateResults();
        }
    }

    private updateResults() {
        if (this.logComponent !== undefined && this._remousResults !== undefined) {
            this.logComponent.log = this._remousResults.log;
            this.generateGraph();
            return true;
        }
        return false;
    }

    private get abscisseIterator(): NumberIterator {
        return this._remousResults.varResults.variatedParameter.paramDefinition.paramValues.getValuesIterator();
    }

    private connectRessaut(lineFlu: LineData, lineTor: LineData) {
        if (lineFlu !== undefined && lineTor !== undefined) {
            const tX = lineFlu.tx.slice(0); // copie

            tX.sort((a, b) => {
                if (a > b) {
                    return 1;
                }
                if (a < b) {
                    return -1;
                }
                return 0;
            });

            let minXflu; // abscisse de début de la courbe fluviale
            let itX = this.abscisseIterator;
            for (const re of this._remousResults.result.resultElements) {
                if (!itX.hasNext) {
                    throw new Error("RemousResultsComponent.connectRessaut() : erreur interne (itérateur sur x)");
                }
                const x = itX.next().value;
                if (re.getExtraResult("flu") !== undefined) {
                    minXflu = x;
                    break;
                }
            }

            if (minXflu !== undefined && minXflu !== tX[0]) {
                // la courbe fluviale ne démarre pas au début, on ajoute un point de raccord avec la ligne torrentielle

                const i = tX.indexOf(minXflu);
                const xflu = tX[i - 1];
                const yflu = lineTor.getYat(xflu);
                lineFlu.setPoint(xflu, yflu);
            }

            let maxXtor; // abscisse de fin de la courbe torrentielle
            const itRE = new ArrayReverseIterator<ResultElement>(this._remousResults.result.resultElements);
            itX = this.abscisseIterator;
            for (const r of itRE) {
                if (!itX.hasNext) {
                    throw new Error("RemousResultsComponent.connectRessaut() : erreur interne (itérateur sur x)");
                }
                const x = itX.next();
                if (r.getExtraResult("tor") !== undefined) {
                    maxXtor = x;
                    break;
                }
            }

            if (maxXtor !== undefined && maxXtor !== tX[tX.length - 1]) {
                // la courbe torrentielle ne finit pas à la fin des abscisses, on ajoute un point de raccord avec la ligne fluviale

                const i = tX.indexOf(maxXtor);
                const xflu = tX[i + 1];
                const yflu = lineFlu.getYat(xflu);
                lineTor.setPoint(xflu, yflu);
            }
        }
    }

    private generateGraph() {
        // http://www.chartjs.org/docs/latest/charts/line.html
        // le dernier dataset de la liste datasets est dessiné en 1er

        this._remousResults.update();
        if (this.varResultsComponent) {
            this.varResultsComponent.results = this._remousResults.varResults;
        }

        const penteFond: number = this._remousResults.penteFond;

        // abscisses

        let labs: number[] = [];
        let xmax: number;
        if (this._remousResults.result.ok) {
            xmax = -1e8;
            const itX2 = this.abscisseIterator;
            while (itX2.hasNext) {
                const x = itX2.next().value;
                labs.push(x);
                xmax = Math.max(x, xmax);
            }
        } else {
            labs = [0, 1];
            xmax = 1;
        }

        // init graphiques

        const gr1 = new GraphData(labs, penteFond, xmax);
        let gr2: GraphData;
        if (this._remousResults.extraGraph) {
            gr2 = new GraphData(labs, 0, xmax);
        }

        // ligne de fond
        gr1.drawLine(0, 0, 3, "#753F00", this.uitextFond, "#753F00");

        // ligne de berge
        if (this._remousResults.hautBerge) {
            gr1.drawLine(this._remousResults.hautBerge, this._remousResults.hautBerge, 4, "#C58F50", this.uitextBerge);
        }

        // hauteur normale
        if (this._remousResults.hautNormale !== undefined && this._remousResults.hautNormale.ok) {
            gr1.drawLine(this._remousResults.hautNormale.vCalc, this._remousResults.hautNormale.vCalc,
                5, "#A4C537", this.uitextTirantNormal
            );
        }

        // hauteur critique
        if (this._remousResults.hautCritique !== undefined && this._remousResults.hautCritique.ok) {
            gr1.drawLine(this._remousResults.hautCritique.vCalc, this._remousResults.hautCritique.vCalc,
                6, "#FF0000", this.uitextTirantCritique
            );
        }

        // lignes d'eau torrentielle et fluviale

        let lineFlu: LineData;
        if (this._remousResults.hasFluData) {
            lineFlu = gr1.newLine(0);
        }
        let lineTor: LineData;
        if (this._remousResults.hasTorData) {
            lineTor = gr1.newLine(1);
        }
        let lineExtra: LineData;
        if (this._remousResults.hasExtra) {
            if (this._remousResults.extraGraph) {
                lineExtra = gr2.newLine(2);
            } else {
                lineExtra = gr1.newLine(2);
            }
        }

        const itX = this.abscisseIterator;
        for (const re of this._remousResults.result.resultElements) {
            if (!itX.hasNext)
                throw new Error("RemousResultsComponent.generateGraph() : erreur interne (itérateur sur x)");

            const x = itX.next().value;
            const yExtra = re.getExtraResult(this._remousResults.extraParamSymbol);
            if (yExtra !== undefined)
                lineExtra.mapPoint(x, yExtra);

            const yFlu = re.getExtraResult("flu");
            if (yFlu !== undefined)
                lineFlu.mapPoint(x, yFlu);

            const yTor = re.getExtraResult("tor");
            if (yTor !== undefined)
                lineTor.mapPoint(x, yTor);
        }

        if (this._remousResults.hasExtra) {
            if (this._remousResults.extraGraph) {
                lineExtra.data = {
                    label: this.extraParamLabel,
                    tension: 0, spanGaps: true, borderColor: "#0093BD", pointRadius: 4
                };
            } else {
                lineExtra.data = {
                    label: this.extraParamLabel,
                    tension: 0, fill: false, spanGaps: true, borderColor: "#C17AF0", pointRadius: 4
                };
            }
        }

        // raccordement ligne fluviale -> torrentielle pour dessiner le ressaut

        this.connectRessaut(lineFlu, lineTor);

        // ajout des données au graphique

        if (lineTor !== undefined)
            lineTor.data = { label: this.uitextLigneTorrentielle, tension: 0, borderColor: "#77A3CD", pointBackgroundColor: "#77A3CD", pointRadius: 4, backgroundColor: "#D1D0D4" };
        if (lineFlu !== undefined)
            lineFlu.data = { label: this.uitextLigneFluviale, tension: 0, borderColor: "#0093BD", pointBackgroundColor: "#0093BD", pointRadius: 4, backgroundColor: "#D1D0D4" };

        this.graph1_data = gr1.data;

        this.graph1_options = {
            responsive: true,
            maintainAspectRatio: true,
            animation: {
                duration: 0
            },
            legend: {
                display: true,
                position: "right"
            },
            title: {
                display: true,
                text: this.uitextAbscisse,
                position: "bottom"
            }
        };

        if (this._remousResults.extraGraph) {
            this.graph2_data = gr2.data;

            this.graph2_options = {
                responsive: true,
                maintainAspectRatio: true,
                animation: {
                    duration: 0
                },
                legend: {
                    display: true,
                    position: "right"
                },
                title: {
                    display: true,
                    text: this.uitextAbscisse,
                    position: "bottom"
                }
            };
        }
    }

    private headers(): string[] {
        return this._tableHeaders;
    }

    private getResultClass(i: number) {
        return "result_id_" + String(i & 1);
    }

    private get hasResults(): boolean {
        return this._remousResults !== undefined && this._remousResults.hasResults;
    }

    private get hasData(): boolean {
        return this._remousResults && this._remousResults.result && this._remousResults.result.ok;
    }
}

/**
 * données pour une ligne dans le graphe
 */
class LineData {
    /**
     * abscisses
     */
    private _tx: number[];

    /**
     * orodonnées
     */
    private _ty: number[] = [];

    /**
     * graphe auquel aapartient la ligne
     */
    private _parentGraph: GraphData;

    /**
     * données fournies à ChartJS
     */
    private _data = {};

    /**
     * profondeur à laquelle est dessinée la ligne
     * les profondeurs les plus petites sont dessinées derrière les profondeurs les plus grandes
     */
    public z: number;

    constructor(gr: GraphData) {
        this._parentGraph = gr;
        this._tx = gr.tx;
        for (let i = this._tx.length - 1; i >= 0; i--)
            this._ty.push(null);
    }

    public getYat(x: number) {
        const i = this._tx.indexOf(x);
        return this._ty[i];
    }

    public setPoint(x: number, y: number) {
        const i = this._tx.indexOf(x);
        this._ty[i] = y;
    }

    public mapPoint(x: number, y: number) {
        this.setPoint(x, this._parentGraph.mapY(x, y));
    }

    public get ty() {
        return this._ty;
    }

    public get tx() {
        return this._tx;
    }

    public hasYs(): boolean {
        for (const y of this._ty)
            if (y != null)
                return true;
        return false;
    }

    public get data() {
        return this._data;
    }

    public set data(d: {}) {
        this._data = d;
        this._data["data"] = this._ty;
    }
}

class GraphData {
    /**
     * tableau des (labels) des abscisses
     */
    private _tx: number[];

    private _lines: LineData[] = [];

    /**
     * pente du fond
    */
    private _penteFond: number;

    /**
     * longueur du bief
     */
    private _longBief: number;

    constructor(tX: number[], pente: number, bief: number) {
        this._tx = tX;
        this._penteFond = pente;
        this._longBief = bief;
    }

    /**
     * crée une ligne dans le graphe
     * @param z profondeur de la lign
     */
    public newLine(z: number): LineData {
        const res = new LineData(this);
        res.z = z;
        this._lines.push(res);
        return res;
    }

    public get tx(): number[] {
        return this._tx;
    }

    /**
   * transforme une cote en tenant compte du fond
   * @param x abscisse où se trouve la cote à transformer
   * @param y cote à transformer
   */
    public mapY(x: number, y: number) {
        let d: number;
        if (this._penteFond >= 0) {
            d = this._penteFond * (this._longBief - x);
        } else {
            d = -this._penteFond * x;
        }
        return y + d;
    }

    /**
     * dessine une ligne droite
     * @param y0 y en x=0
     * @param ymax y en x=xmax
     * @param color couleur de la ligne
     * @param lbl légende de la ligne
     * @param fillColor couleur de remplissage sous la ligne
     */
    public drawLine(y0: number, ymax: number, prof: number, color: string, lbl: string, fillColor: string = undefined) {
        const l = this.newLine(prof);
        l.mapPoint(0, y0);
        l.mapPoint(this._longBief, ymax);

        // l.data = { label: lbl, data: l, fill: fillColor != undefined, tension: 0, borderColor: color, backgroundColor: fillColor, pointRadius: 0 };
        l.data = {
            label: lbl, fill: fillColor !== undefined, tension: 0, spanGaps: true,
            borderColor: color, backgroundColor: fillColor, pointRadius: 0
        };
    }

    public get data() {
        const ds = [];
        this._lines.sort((a, b) => {
            if (a.z > b.z)
                return -1;
            if (a.z < b.z)
                return 1;
            return 0;
        });

        for (const l of this._lines)
            ds.push(l.data);

        return {
            labels: this._tx,
            datasets: ds
        };
    }
}
